import json
import logging
from multiprocessing import Process
from multiprocessing import Queue
from queue import Empty

import requests

try:
    from scribereader import scribereader
except ImportError:
    scribereader = None

SLACK_WEBHOOK_STREAM = 'stream_slack_incoming_webhook'
SCRIBE_ENV = 'uswest1-prod'
log = logging.getLogger(__name__)


def get_slack_blocks_for_deployment(
    deployment_name: str,
    message,
    last_action=None,
    status=None,
    progress=None,
    active_button=None,
    available_buttons=["rollback", "forward"],
    from_sha=None,
    to_sha=None,
    slo_watchers=None,
):

    button_elements = get_button_elements(
        available_buttons, active_button=active_button,
        from_sha=from_sha, to_sha=to_sha,
    )
    blocks = [
        {
            "type": "section",
            "text": {
                "type": "mrkdwn",
                "text": f"{deployment_name}",
            },
        },
        {
            "type": "section",
            "text": {
                "type": "mrkdwn",
                "text": message,
            },
        },
        {
            "type": "section",
            "text": {
                "type": "mrkdwn",
                "text": f"State machine: `{status}`\nProgress: {progress}\nLast operator action: {last_action}",
            },
        },
    ]
    if slo_watchers is not None and len(slo_watchers) > 0:
        num_failing = len([w for w in slo_watchers if w.failing])
        if num_failing > 0:
            blocks.append({
                "type": "section",
                "text": {
                    "type": "mrkdwn",
                    "text": f":alert: {num_failing} of {len(slo_watchers)} SLOs are failing",
                },
            })
        else:
            blocks.append({
                "type": "section",
                "text": {
                    "type": "mrkdwn",
                    "text": f":ok_hand: All {len(slo_watchers)} SLOs are currently passing.",
                },
            })
    else:
        blocks.append({
            "type": "section",
            "text": {
                "type": "mrkdwn",
                "text": "No SLOs defined for this service.",
            },
        })

    if button_elements != []:
        blocks.append({
            "type": "actions",
            "block_id": "deployment_actions",
            "elements": button_elements,
        })
    return blocks


def get_button_element(button, is_active, from_sha, to_sha):
    active_button_texts = {
        "rollback": f"Rolling Back to {from_sha[:8]} :zombocom:",
        "forward": f"Rolling Forward to {to_sha[:8]} :zombocom:",
    }

    inactive_button_texts = {
        "rollback": f"Roll Back to {from_sha[:8]} :arrow_backward:",
        "forward": f"Continue Forward to {to_sha[:8]} :arrow_forward:",
        "complete": f"Complete deploy to {to_sha[:8]} :white_check_mark:",
        "abandon": f"Abandon deploy, staying on {from_sha[:8]} :x:",
        "snooze": f"Reset countdown",
    }

    if is_active is True:
        confirm = False
        text = active_button_texts[button]
    else:
        confirm = get_confirmation_object(button)
        text = inactive_button_texts[button]

    element = {
        "type": "button",
        "text": {
            "type": "plain_text",
            "text": text,
            "emoji": True,
        },
        "confirm": confirm,
        "value": button,
    }
    if not confirm:
        del element["confirm"]
    return element


def get_button_elements(buttons, active_button=None, from_sha=None, to_sha=None):
    elements = []
    for button in buttons:
        is_active = button == active_button
        elements.append(
            get_button_element(button=button, is_active=is_active, from_sha=from_sha, to_sha=to_sha),
        )
    return elements


def get_confirmation_object(action):
    return {
        "title": {
            "type": "plain_text",
            "text": "Are you sure?",
        },
        "text": {
            "type": "mrkdwn",
            "text": f"Did you mean to press {action}?",
        },
        "confirm": {
            "type": "plain_text",
            "text": "Yes. Do it!",
        },
        "deny": {
            "type": "plain_text",
            "text": "Stop, I've changed my mind!",
        },
    }


class ButtonPress():
    def __init__(self, event):
        self.event = event
        self.username = event["user"]["username"]
        self.response_url = event["response_url"]
        # TODO: Handle multiple actions?
        self.action = event["actions"][0]["value"]
        self.thread_ts = event["container"].get("thread_ts", None)
        self.channel = event["channel"]["name"]

    def __repr__(self):
        return self.event

    def update(self, blocks):
        # Implements responding to button presses
        # https://api.slack.com/messaging/interactivity/enabling#responding-to-interactions
        # But isn't the api_call method per-se
        # https://github.com/slackapi/python-slackclient/issues/270
        requests.post(self.response_url, json={"blocks": blocks})


def event_to_buttonpress(event):
    return ButtonPress(event=event)


def parse_webhook_event_json(line):
    event = json.loads(line)
    log.debug(event)
    return event


def is_relevant_event(event):
    # TODO: Implement filtering
    return True


def get_slack_events():
    if scribereader is None:
        logging.error("Scribereader unavailable. Not tailing slack events.")
        return

    def scribe_tail(queue):
        host_and_port = scribereader.get_env_scribe_host(SCRIBE_ENV, True)
        host = host_and_port['host']
        port = host_and_port['port']
        tailer = scribereader.get_stream_tailer(SLACK_WEBHOOK_STREAM, host, port)
        for line in tailer:
            queue.put(line)

    # Tailing scribe is not thread-safe, therefore we must use a Multiprocess-Queue-based
    # approach, with paasta logs as prior art.
    queue = Queue()
    kw = {'queue': queue}
    process = Process(target=scribe_tail, daemon=True, kwargs=kw)
    process.start()
    while True:
        try:
            line = queue.get(block=True, timeout=0.1)
            event = parse_webhook_event_json(line)
            if is_relevant_event(event):
                yield event
        except Empty:
            pass



import functools
import os
import textwrap
import threading
import time
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple

import pytimeparse
try:
    from slo_utils.yelpsoa_configs import get_slo_files_from_soaconfigs
    from slo_transcoder.composite_sinks.signalform_detectors import make_signalform_detector_composite_sink
    from slo_transcoder.sources.yelpsoaconfigs import YelpSoaConfigsSource
    SLO_TRANSCODER_LOADED = True
except ImportError:
    SLO_TRANSCODER_LOADED = False

import tempfile

from paasta_tools.utils import DEFAULT_SOA_DIR
from paasta_tools.automatic_rollbacks.signalfx import tail_signalfx


def get_slos_for_service(service, soa_dir=DEFAULT_SOA_DIR):
    if not SLO_TRANSCODER_LOADED:
        return

    slo_files = [f for f in get_slo_files_from_soaconfigs(soa_dir) if f.startswith(os.path.join(soa_dir, service))]
    with tempfile.TemporaryDirectory() as temp_dir:
        SFDCS = make_signalform_detector_composite_sink(output_directory=temp_dir)
        sources = [YelpSoaConfigsSource(file) for file in slo_files]
        composite_sink = SFDCS(service=service, sources=sources)

        for sink in composite_sink.sinks():
            signalflow, rules = sink.generate_signalflow_signals_and_rules()
            query = textwrap.dedent('\n'.join(signalflow))
            yield sink, query


class SLODemultiplexer:
    def __init__(
        self,
        sink,
        individual_slo_callback: Callable[[str, bool], None],
        all_slos_callback: Callable[[bool], None],
        start_timestamp: Optional[float] = None,
        max_duration: float = 300,
    ) -> None:
        self.sink = sink
        self.individual_slo_callback = individual_slo_callback
        self.all_slos_callback = all_slos_callback
        self.max_duration = max_duration
        if start_timestamp is None:
            self.start_timestamp = time.time()
        else:
            self.start_timestamp = start_timestamp

        self.slo_watchers_by_label: Dict[str, 'SLOWatcher'] = {}
        for slo in sink.slos:
            label = sink._get_detector_label(slo)
            watcher = SLOWatcher(slo, self.watcher_callback, self.start_timestamp, label, max_duration=max_duration)
            self.slo_watchers_by_label[label] = watcher

        self.failing = False

    def process_datapoint(self, props, datapoint, timestamp) -> None:
        slo_label = props['dimensions']['sf_metric'].rsplit('.', 1)[0]
        watcher = self.slo_watchers_by_label[slo_label]
        watcher.process_datapoint(props, datapoint, timestamp)

    def watcher_callback(self, watcher: 'SLOWatcher') -> None:
        old_failing = self.failing
        new_failing = any(watcher.failing for watcher in self.slo_watchers_by_label.values())
        self.individual_slo_callback(watcher.label, watcher.failing)

        self.failing = new_failing
        if new_failing == (not old_failing):
            self.all_slos_callback(new_failing)


class SLOWatcher:
    def __init__(self, slo, callback, start_timestamp, label, max_duration=300):
        self.slo = slo
        self.window = []
        self.max_duration = max_duration
        self.failing = None
        self.bad_after_mark = None
        self.bad_before_mark = None
        self.callback = callback
        self.start_timestamp = start_timestamp
        self.label = label

    def process_datapoint(self, props, datapoint, timestamp) -> None:
        self.window.append((timestamp, datapoint))
        self.trim_window()

        if timestamp > self.start_timestamp:
            new_bad_after_mark = self.is_window_bad()
            new_bad_before_mark = self.bad_before_mark
        else:
            new_bad_after_mark = self.bad_after_mark
            new_bad_before_mark = self.is_window_bad()

        new_failing = new_bad_after_mark and not new_bad_before_mark
        old_failing = self.failing

        self.bad_after_mark = new_bad_after_mark
        self.bad_before_mark = new_bad_before_mark
        self.failing = new_failing

        if new_failing == (not old_failing):
            self.callback(self)

    def trim_window(self) -> None:
        old_window = self.window
        min_ts = max(ts for ts, d in old_window) - self.window_duration()
        new_window = [(ts, d) for (ts, d) in old_window if ts > min_ts]
        self.window = new_window

    def window_duration(self) -> float:
        """Many of our SLOs are defined with durations of 1 hour or more; this is great if you're trying to avoid being
        paged, but not helpful for a deployment that you expect to finish within a few minutes. self.max_duration allows
        us to cap the length of time that we consider. This should make us a bit more sensitive."""
        return min(self.max_duration, pytimeparse.parse(str(self.slo.config.duration)))

    def is_window_bad(self) -> bool:
        bad_datapoints = len([1 for ts, d in self.window if d > self.slo.config.threshold])
        return (bad_datapoints / len(self.window)) >= 1. - self.slo.config.percent_of_duration / 100.


def watch_slos_for_service(
    service: str,
    individual_slo_callback: Callable[[str, bool], Any],
    all_slos_callback: Callable[[bool], Any],
    sfx_api_token: str,
) -> Tuple[List[threading.Thread], List[SLODemultiplexer]]:
    threads = []
    watchers = []

    for sink, query in get_slos_for_service(service):
        watcher = SLODemultiplexer(
            sink,
            individual_slo_callback=individual_slo_callback,
            all_slos_callback=all_slos_callback,
        )
        thread = threading.Thread(
            target=functools.partial(
                tail_signalfx,
                query,
                lookback_seconds=30,
                callback=watcher.process_datapoint,
                sfx_api_token=sfx_api_token,
            ),
            daemon=True,
        )
        threads.append(thread)
        thread.start()
        watchers.append(watcher)

    return threads, watchers



# Copyright 2015-2019 Yelp Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any
from typing import List
from typing import Mapping
from typing import Optional

from kubernetes.client.rest import ApiException

from paasta_tools.kubernetes_tools import KubeClient
from paasta_tools.utils import DEFAULT_SOA_DIR
from paasta_tools.utils import InstanceConfig


FLINK_INGRESS_PORT = 31080


class FlinkConfig(InstanceConfig):
    config_filename_prefix = 'flink'

    def __init__(self, service, instance, cluster, soa_dir=DEFAULT_SOA_DIR):
        super().__init__(
            cluster=cluster,
            instance=instance,
            service=service,
            soa_dir=soa_dir,
            config_dict={},
            branch_dict={},
        )

    def validate(self) -> List[str]:
        # Use InstanceConfig to validate shared config keys like cpus and mem
        error_msgs = super().validate()

        if error_msgs:
            name = self.get_instance()
            return [f'{name}: {msg}' for msg in error_msgs]
        else:
            return []


def load_flink_instance_config(
    service: str,
    instance: str,
    cluster: str,
    load_deployments: bool = True,
    soa_dir: str = DEFAULT_SOA_DIR,
) -> FlinkConfig:
    return FlinkConfig(
        service=service,
        instance=instance,
        cluster=cluster,
    )


def sanitised_name(
    service: str,
    instance: str,
) -> str:
    sanitised_service = service.replace('_', '--')
    sanitised_instance = instance.replace('_', '--')
    return f'{sanitised_service}-{sanitised_instance}'


def flink_custom_object_id(service: str, instance: str) -> Mapping[str, str]:
    return dict(
        group='yelp.com',
        version='v1alpha1',
        namespace='paasta-flinks',
        plural='flinks',
        name=sanitised_name(service, instance),
    )


def get_flink_config(
    kube_client: KubeClient,
    service: str,
    instance: str,
) -> Optional[Mapping[str, Any]]:
    try:
        co = kube_client.custom.get_namespaced_custom_object(
            **flink_custom_object_id(service, instance),
        )
        status = co.get('status')
        return status
    except ApiException as e:
        if e.status == 404:
            return None
        else:
            raise


def set_flink_desired_state(
    kube_client: KubeClient,
    service: str,
    instance: str,
    desired_state: str,
):
    co_id = flink_custom_object_id(service, instance)
    co = kube_client.custom.get_namespaced_custom_object(**co_id)
    if co.get('status', {}).get('state') == desired_state:
        return co['status']

    if 'metadata' not in co:
        co['metadata'] = {}
    if 'annotations' not in co['metadata']:
        co['metadata']['annotations'] = {}
    co['metadata']['annotations']['yelp.com/desired_state'] = desired_state
    kube_client.custom.replace_namespaced_custom_object(**co_id, body=co)
    status = co.get('status')
    return status


def get_dashboard_url(
    cluster: str,
    service: str,
    instance: str,
) -> str:
    sname = sanitised_name(service, instance)
    return f'http://flink.k8s.paasta-{cluster}.yelp:{FLINK_INGRESS_PORT}/{sname}'
